#######################################################################
#                                Types                                #
#######################################################################


struct BernsteinReconstruction{T_Mesh<:Mesh1d} <: AbstractReconstruction
  mesh::T_Mesh
  P::Matrix # projection matrix to transfer data onto Bernstein nodes
  V::Matrix # vandermonde matrix to resum Bernstein series on nodes

  function BernsteinReconstruction(mesh, P, I)
    # reconstruction should maintain grid structure
    @assert size(P)[1] == size(I)[2]
    @assert size(P)[2] == size(I)[1]
    return new{typeof(mesh)}(mesh, P, I)
  end
end


#######################################################################
#                            Constructors                             #
#######################################################################


function BernsteinReconstruction(mesh::Mesh)
  z = mesh.element.z
  Npts = length(z)
  rng = collect(0:Npts-1)
  bernstein_z = @. -1 + 2 * rng / (Npts-1)
  P = dg1d.lagrange_interp_matrix(bernstein_z, z)
  V = vandermonde_matrix_bernstein(z)
  return BernsteinReconstruction(mesh, P, V)
end


#######################################################################
#                               Methods                               #
#######################################################################


# Reference contains a typo in the normalization of the polynomials,
# it should be n instead of N.
function bernstein_polynomial(x, n, N)
  return binomial(N, n) / 2^N * (1 + x)^n * (1 - x)^(N-n)
end


function vandermonde_matrix_bernstein(z)
  Npts = length(z)
  V = [ bernstein_polynomial(z[i], j-1, Npts-1) for i=1:Npts, j=1:Npts ]
  return V
end


@doc """
    reconstruct_cell!(u!, w, P, V; kwargs...)

Compute inplace a convex combination of `u!` and its Bernstein reconstruction weighted by `w`.
`P, V` are the project and transfer matrices that transfer data from the sample
grid to the bernstein grid and back, respectively.
`m` and `M` keywords allow to enforce lower and upper bound in reconstruction.
"""
@inline function reconstruct_cell!(u, w, P, V; m=missing, M=missing)

  uB = P * u

  if !ismissing(m)
    @. uB = max(uB, m)
  end
  if !ismissing(M)
    @. uB = min(uB, M)
  end

  u .*= w
  u .+= (1-w) * V * uB
  return
end


"""
    function reconstruct!(u, weights, bernstein::BernsteinReconstruction;
        m=missing, M=missing, isperiodic::Bool=true, limit_with_boundaries::Bool=true)

Apply `bernstein` reconstruction to nodal values `u` inplace for each cell based using
troubled cell weights `w`. Use `m, M` to impose lower and upper bounds on the result.
Set `limit_with_boundaries=true` to use each cells boundary values as lower and upper bounds
(they are not altered by the reconstruction). Using `limit_with_boundaries=true` and supplying
`m,M` amounts to `m = max(min(ul,ur), m), M = min(max(ul,ur), M)` for every cell, where
`ul, ur` are the cell's left and right boundary values.
Option `isperiodic` is unused.
"""
function reconstruct!(u, weights, bernstein::BernsteinReconstruction;
    m=missing, M=missing, isperiodic::Bool=true, limit_with_boundaries::Bool=true)

  @unpack mesh, P, V = bernstein
  L = layout(mesh)
  Npts, K = L

  u = reshape(view(u,:), L)
  for k = 1:K
    w = weights[k]
    cell_u = view(u, :, k)
    mm, MM = if limit_with_boundaries
      ul, ur = first(cell_u), last(cell_u)
      minu, maxu = min(ul, ur), max(ul, ur)
      mm = ismissing(m) ? minu : max(minu, m)
      MM = ismissing(M) ? maxu : min(maxu, M)
      mm, MM
    else
      m, M
    end
    reconstruct_cell!(cell_u, 1-w, P, V, m=mm, M=MM)
  end

end
