function make_callback(env, P::Project, isperiodic::Bool)
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic, P.equation, env.mesh)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic, P.tci, env.mesh)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic, P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cb_step, cb_substep = if get(P.prms, :av_recompute_substeps, false) && P.hrsc isa AbstractArtificialViscosity
    CallbackSet(cb_equation, cb_tci), CallbackSet(cb_hrsc)
  else
    CallbackSet(cb_equation, cb_tci, cb_hrsc), nothing
  end
  cbfn_analyze(u, t)  = callback_analyze(u, t, env, P, env.mesh)
  cb_analyze          = FunctionCallback(cbfn_analyze,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  push!(cb_step, cb_analyze)
  if P.prmsdb["ScalarEq"]["analyze_error"]
    cbfn_analyze_error(u, t)  = callback_analyze_error(u, t, env, P, env.mesh)
    cb_analyze_error          = FunctionCallback(cbfn_analyze_error,
                                                 CallbackTiming(every_iteration=1,every_dt=0))
    push!(cb_step, cb_analyze_error)
  end
  return cb_step, cb_substep
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P, isperiodic, eq::AbstractScalarEq, ::Mesh) end


function callback_equation(state_u, state_t, env, P, isperiodic, eq::AbstractScalarEq, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P
  wrap_dynamic_variables!(cache, state_u)
  @unpack u                     = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1        = get_static_variables(cache)
  @unpack t, tm1                = get_global_variables(cache)

  # backup previous values
  @. Em1  = E
  @. Fm1  = F
  @. tm1  = t
  t      .= state_t
  # compute new values
  broadcast_volume!(entropy, equation, mesh)
  broadcast_volume!(entropy_flux, equation, mesh)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Nothing, mesh) = nothing


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractHRSC, mesh) =
  error("not yet implemented for type $(typeof(hrsc))")


function callback_hrsc(state_u, state_t, env, P, isperiodic,
  hrsc::Union{HRSC.BernsteinReconstruction, HRSC.WENOReconstruction}, mesh) end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractArtificialViscosity, mesh)
  @unpack cache, mesh = env
  @unpack equation = P
  @unpack v         = get_static_variables(cache)
  @unpack mu, max_v = get_cell_variables(cache)
  @unpack u         = get_dynamic_variables(cache)
  broadcast_volume!(speed, equation, mesh)
  Npts, K = layout(mesh)
  # max_v .= dg1d.cellwise_average(v, mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.TCIViscosity, mesh)
  @unpack cache, mesh = env
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh1d)
  @unpack mu, mu_m1 = get_cell_variables(mesh)
  mu_m1 .= mu
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack av_drag = P.prms
  @. mu = ((1-av_drag)*mu + av_drag*mu_m1)
  @unpack cache = env
  @unpack smoothed_mu = get_static_variables(cache)
  hrsc.smoother(smoothed_mu,mu,mesh,isperiodic)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh2d)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache = env
  @unpack mu = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  hrsc.smoother(smoothed_mu,mu,mesh,true,true)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P
  @unpack u              = get_dynamic_variables(cache)
  @unpack E, F, Em1, Fm1, v = get_static_variables(cache)
  @unpack mu, max_v      = get_cell_variables(cache)
  @unpack t, tm1         = get_global_variables(cache)
  broadcast_volume!(speed, equation, mesh)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v, E, Em1, F, Fm1, t[1], tm1[1],
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh2d)
  @unpack cache, mesh  = env
  @unpack equation     = P
  @unpack u            = get_dynamic_variables(cache)
  @unpack E, Em1, v,
          F_x, F_y,
          F_xm1, F_ym1,
          dxdX, dydY   = get_static_variables(cache)
  @unpack t, tm1       = get_global_variables(cache)
  @unpack mu           = get_cell_variables(cache)

  # backup previous values
  @. Em1   = E
  @. F_xm1 = F_x
  @. F_ym1 = F_y
  @. tm1   = t
  @. t     = state_t

  # compute new values
  broadcast_volume!(entropy, equation, mesh)
  broadcast_volume!(speed, equation, mesh)
  broadcast_volume!(entropy_flux, equation, mesh)

  HRSC.compute_viscosity!(
      mu,
      u, v, E, Em1, F_x, F_y, F_xm1, F_ym1, t[1], tm1[1], dxdX, dydY,
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


callback_tci(state_u, state_t, env, P, isperiodic, tci::Nothing, mesh) = nothing


callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.AbstractTCI) =
  error("not yet implemented for type $(typeof(tci))")


  function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.EntropyProduction, ::Mesh1d)
  @unpack cache, mesh            = env
  @unpack u                      = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1, EP     = get_static_variables(cache)
  @unpack flag                   = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @unpack t, tm1                 = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, F, Fm1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, env, P, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest}, mesh)
  @unpack cache, mesh = env
  @unpack u           = get_dynamic_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  TCI.compute_indicator!(flag, u, tci)
end


#######################################################################
#                            analyze_error                            #
#######################################################################


callback_analyze_error(state_u, state_t, env, P, mesh::Mesh1d{FVElement}) = TODO()


function callback_analyze_error(state_u, state_t, env, P, mesh::Mesh1d)
  @unpack u = get_dynamic_variables(mesh)
  @unpack u_err, u_ana = get_static_variables(mesh)
  compute_analytic_ref_solution!(mesh, P.equation, P.prmsdb, state_t)
  @. u_err = u - u_ana
  if P.prms.analyze_error_norm
    @unpack u_err_norm = get_global_variables(mesh)
    u_err_norm .= dg1d.norm_L2(u_err, mesh)
  end
end


compute_analytic_ref_solution!(mesh, eq, prms, t) = TODO(eq)


function compute_analytic_ref_solution!(mesh::Mesh1d, eq::Advection, prms, t)
  @unpack u_ana, x = get_static_variables(mesh)
  xmin, xmax = mesh.extends[1]
  L = xmax-xmin
  if prms["ScalarEq"]["id"] === "sine"
    id_sine_period = prms["ScalarEq"]["id_sine_period"]
    id_sine_phase = prms["ScalarEq"]["id_sine_phase"]
    @unpack v = eq
    u_ana .= @. sin.(2 * pi * id_sine_period * (x-v*t)/L + id_sine_phase)
  else
    TODO(P.prms.id)
  end
end


#######################################################################
#                               analyze                               #
#######################################################################


# TODO
callback_analyze(state_u, state_t, env, P, mesh::Mesh) = nothing


function callback_analyze(state_u, state_t, env, P, mesh::Mesh1d)
  @unpack u = get_dynamic_variables(mesh)
  @unpack u_norm2 = get_global_variables(mesh)
  u_norm2 .= dg1d.norm_L2(u, mesh)^2
end
function callback_analyze(state_u, state_t, env, P, mesh::Mesh1d{<:FVElement})
  @unpack u = get_dynamic_variables(mesh)
  @unpack u_norm2 = get_global_variables(mesh)
  if P.prms.hrsc === :fv_central
    @unpack δu = get_static_variables(mesh)
    prms = (hrsc = :fv_central,)
    # TODO need to limit slope here, because rhs! does not do it (yet?)
    dg1d.limit_slope!(δu, u, mesh, prms)
    u_norm2 .= dg1d.integrate_fv_central(ui -> ui^2, δu, u, mesh)
  elseif P.prms.hrsc === :slope_limiter
    @unpack uL, uR = get_static_variables(mesh)
    limit_slope!(P, mesh, state_u)
    u_norm2 .= dg1d.integrate_fv_slopelimited(ui -> ui^2, uL, uR, mesh)
  else
    TODO(P.prms.hrsc)
  end
end
