# TODO Rename to DGElement
struct SpectralElement
  N::Int64                # polynomial order
  Npts::Int64             # number of quadrature points = N + 1
  kind::Symbol            # kind of nodal approximation

  # quadrature rule
  quadr::Symbol            # :LGL, :LGL_mass_lumped, :GLGL
  quadr_z::Vector{Float64} # quadrature points
  quadr_w::Vector{Float64} # quadrature weights
  quadr_v::Float64         # derivative weight (only needed for GLGL quadrature rule)
  quadr_D::Matrix{Float64} # derivative matrix based on quadr_z

  # nodal points and DG operators
  z::Vector{Float64}      # nodal points
  D::Matrix{Float64}      # derivative matrix based on z
  S::Matrix{Float64}      # stiffness matrix
  invM::Matrix{Float64}   # inverse mass matrix
  MDM::Matrix{Float64}    # = (M^-1 D^T M) = M^-1 S^T
  # TODO Rename to invMϕ_lhs, invMϕ_rhs
  Ml_lhs::Vector{Float64} # = M^-1 l(-1)
  Ml_rhs::Vector{Float64} # = M^-1 l(+1)

  # helpers
  V::Matrix{Float64}      # = P_(j-1)(z_i), Vandermonde matrix
  invV::Matrix{Float64}   # inv(V), inverse Vandermonde matrix
  F::Matrix{Float64}      # spectral filtering matrix

  function SpectralElement(N::Int, quadr::Symbol, kind::Symbol;
      # filter parameters
      ηc::Float64=0.5, sh::Float64=16.0, α::Float64=36.0)

    Npts = N + 1

    if kind === :nodal_legendre
      if quadr === :LGL
        qaudr_z, qaudr_w = LGL.rule(Npts); qaudr_v = -1.0 # unused here
        purge_zeros!(qaudr_z)
        S    = LGL.stiffness_matrix(Npts)
        invM = LGL.inverse_mass_matrix_exact(Npts)
      elseif quadr === :LGL_lumped
        quadr_z, quadr_w = LGL.rule(Npts); quadr_v = -1.0 # unused here
        purge_zeros!(quadr_z)
        S    = LGL.stiffness_matrix(Npts)
        invM = LGL.inverse_mass_matrix_lumped(Npts)
      elseif quadr === :GLGL
        quadr_z, quadr_w, quadr_v = GLGL.rule(Npts)
        purge_zeros!(quadr_z)
        quadr_D  = first_derivative_matrix(quadr_z)
        quadr_D2 = second_derivative_matrix(quadr_z)
        M    = GLGL.mass_matrix(quadr_z, quadr_w, quadr_v, quadr_D)
        S    = GLGL.stiffness_matrix(quadr_z, quadr_w, quadr_v, quadr_D, quadr_D2)
        invM = inv(M)
      else
        error("Invalid quadrature rule specified for kind $kind: $quadr. Use one of :LGL_lumped, :LGL, :GLGL.")
      end
      z = quadr_z
      D = first_derivative_matrix(z)
      quadr_D = D
      l_lhs = zeros(Float64, N+1)
      l_rhs = zeros(Float64, N+1)
      l_lhs[1]   = 1.0
      l_rhs[end] = 1.0
      V = vandermonde_matrix_legendre(z)
    # In preparation for Bernstein basis
    # elseif kind === :nodal_bernstein
    #   if quadr === :LGL || quadr === :LGL_lumped
    #     qaudr_z, qaudr_w = LGL.rule(Npts)
    #     qaudr_v          = -1.0 # unused here
    #     purge_zeros!(qaudr_z)
    #   elseif quadr === :GLGL
    #     quadr_z, quadr_w, quadr_v = GLGL.rule(Npts)
    #     purge_zeros!(quadr_z)
    #   else
    #     error("Invalid quadrature rule specified for kind $kind: $quadr. Use one of :LGL_lumped, :LGL, :GLGL.")
    #   end
    #   z = collect(range(-1.0,1.0,Npts))
    #   purge_zeros!(z)
    #   M = Bernstein.mass_matrix(Npts)
    #   S = Bernstein.stiffness_matrix(Npts)
    #   invM = inv(M)
    #   D = Bernstein.first_derivative_matrix(Npts)
    #   TODO("quadr_D")
    #   TODO("quadr_w")
    #   V  = Bernstein.vandermonde_matrix(z)
    #   I  = stack(Bernstein.bernstein_polynomial.(quadr_z, i-1, N) for i in 1:Npts)
    #   quadr_w = transpose(I) * quadr_w
    #   quadr_D = D * I
    #   l_lhs   = zeros(Float64, N+1)
    #   l_rhs   = zeros(Float64, N+1)
    #   l_lhs[1]    = 1.0
    #   l_rhs[end]  = 1.0
    end

    MDM  = invM * transpose(S); purge_zeros!(MDM)
    invV = inv(V)
    Λ = diagm([ spectral_filter((i-1)/N, (ηc,sh,α)) for i = 1:Npts ])
    F = V * Λ * invV
    Ml_lhs = invM * l_lhs
    Ml_rhs = invM * l_rhs

    return new(N, Npts, kind,
               quadr, quadr_z, quadr_w, quadr_v, quadr_D,
               z, D, S, invM, MDM, Ml_lhs, Ml_rhs,
               V, invV, F)
  end
end


# (5.16) from Hesthaven, Warburton 2007
function spectral_filter(η, prms)
  ηc, sh, α = prms
  return if 0 ≤ η ≤ ηc
    1
  else
    exp(-α*((η-ηc)/(1-ηc))^(2*sh))
  end
end


@inline function inner_product(u1, u2, el::SpectralElement)
  if el.quadr === :LGL || el.quadr === :LGL_lumped
    return LGL.integrate(u1, u2, el.quadr_z, el.quadr_w)
  else # el.quadr === :GLGL
    return GLGL.integrate(u1, u2, el.quadr_z, el.quadr_w, el.quadr_v, el.quadr_D)
  end
end


@inline function integrate(u, el::SpectralElement)
  if el.quadr === :LGL || el.quadr === :LGL_lumped
    return LGL.integrate(u, el.quadr_z, el.quadr_w)
  else # el.quadr === :GLGL
    return GLGL.integrate(u, el.quadr_z, el.quadr_w, el.quadr_v, el.quadr_D)
  end
end


@inline function integrate(fn::Function, idxs::AbstractVector{<:Integer}, el::SpectralElement)
  if el.quadr === :LGL || el.quadr === :LGL_lumped
    return LGL.integrate(fn, idxs, el.quadr_z, el.quadr_w)
  else # el.quadr === :GLGL
    return GLGL.integrate(fn, idxs, el.quadr_z, el.quadr_w, el.quadr_v, el.quadr_D)
  end
end


@inline function differentiate(u, el::SpectralElement)
  return el.D * u
end

@inline function differentiate!(du, u, el::SpectralElement)
  mul!(du, el.D, u)
end
